'use strict';

var _extends2 = require('babel-runtime/helpers/extends');

var _extends3 = _interopRequireDefault(_extends2);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var express = require('express');
var bodyParser = require('body-parser');

var _require = require('graphql'),
    parse = _require.parse,
    print = _require.print;

var _require2 = require('graphql-tools'),
    makeExecutableSchema = _require2.makeExecutableSchema;

var _require3 = require('apollo-server-express'),
    graphqlExpress = _require3.graphqlExpress,
    graphiqlExpress = _require3.graphiqlExpress;

var cors = require('cors');

var _require4 = require('apollo-errors'),
    createError = _require4.createError,
    formatError = _require4.formatError;

var compression = require('compression');

var NextServer = require('../../server/next-plus');

var _require5 = require('../utils'),
    Hook = _require5.Hook,
    HEADER_LOGIN_TOKEN_KEY = _require5.HEADER_LOGIN_TOKEN_KEY,
    LOGIN_TOKEN_KEY = _require5.LOGIN_TOKEN_KEY;

NextServer.ApolloError = createError('ApolloError', {
  message: 'Apollo Error'
});

var configGraphqlOptions = function configGraphqlOptions(func) {
  return this._configGraphqlOptionsHook.register(func);
};

var configGraphqlSchema = function configGraphqlSchema(func) {
  return this._configGraphqlSchemaHook.register(func);
};

NextServer.addServerModule('defaultApolloServerModule', {
  setup: function setup() {
    this._configGraphqlOptionsHook = new Hook();
    this._configGraphqlSchemaHook = new Hook();

    this.configGraphqlOptions = configGraphqlOptions.bind(this);
    this.configGraphqlSchema = configGraphqlSchema.bind(this);
  },
  afterStartup: function afterStartup() {
    this.graphqlServer = createApollolServer(this);
    this.appHandlers.use(this.graphqlServer);
  }
});

function createApollolServer(server) {
  var config = {
    path: server.getSetting('graphqlPath', '/graphql'),
    graphiql: server.getSetting('graphiql', server.dev),
    graphiqlPath: server.getSetting('graphiqlPath', '/graphiql'),
    graphiqlOptions: {
      passHeader: '"' + HEADER_LOGIN_TOKEN_KEY + '": localStorage["' + LOGIN_TOKEN_KEY + '"] || ""'
    }
  };

  server._configGraphqlSchemaHook.each(function (func) {
    func({
      schemas: server.graphqlSchemaObj.schemas,
      queryFields: server.graphqlSchemaObj.queryFields,
      mutationFields: server.graphqlSchemaObj.mutationFields,
      resolvers: server.graphqlResolvers
    });
    return true;
  });

  var schema = generateExecutableSchema({
    schemaObj: server.graphqlSchemaObj,
    resolvers: server.graphqlResolvers
  });
  server.executableSchema = schema;
  server._typeDefs = schema._typeDefs;

  var defaultGraphqlOptions = {
    formatError: function formatError(e) {
      return {
        message: e.message,
        locations: e.locations,
        path: e.path
      };
    },
    debug: server.dev,
    schema: schema,
    context: (0, _extends3.default)({}, server.graphqlContext),
    // apollo engine
    tracing: true,
    cacheControl: true
  };

  var graphqlServer = express();

  var whitelist = [];
  try {
    whitelist.push(server.absoluteUrl().replace(/\/$/, ''));
  } catch (__) {}
  whitelist = whitelist.concat(server.getSetting('graphqlServerCORSWhitelist', []));
  var corsOptions = {
    origin: function origin(_origin, callback) {
      var originIsWhitelisted = whitelist.indexOf(_origin) !== -1;
      if (!originIsWhitelisted) {
        for (var i = 0; i < whitelist.length; i++) {
          var pattern = whitelist[i];
          try {
            var isMatch = new RegExp(pattern).test(_origin);
            if (isMatch) {
              originIsWhitelisted = true;
              break;
            }
          } catch (__) {}
        }
      }
      callback(null, originIsWhitelisted);
    },

    credentials: true,
    optionsSuccessStatus: 200
  };
  graphqlServer.use(cors(corsOptions));

  graphqlServer.use(compression());

  graphqlServer.use(config.path, bodyParser.json(), graphqlExpress(function (req) {
    var options = (0, _extends3.default)({}, defaultGraphqlOptions, {
      context: (0, _extends3.default)({}, defaultGraphqlOptions.context, {
        req: req,
        loginToken: req.headers && req.headers[HEADER_LOGIN_TOKEN_KEY]
      })
    });

    options.formatError = formatError;

    try {
      server._configGraphqlOptionsHook.each(function (func) {
        options = func(options);
        if (!options) {
          throw new Error('get graphql options error');
        }
        return true;
      });
    } catch (error) {
      console.error('> Apollo server error:', error);
    }

    return options;
  }));

  if (config.graphiql) {
    graphqlServer.use(config.graphiqlPath, graphiqlExpress((0, _extends3.default)({}, config.graphiqlOptions, {
      endpointURL: server.getSetting('graphqlEndpointURL', config.path)
    })));
  }

  return graphqlServer;
}

function generateExecutableSchema(_ref) {
  var schemaObj = _ref.schemaObj,
      resolvers = _ref.resolvers;

  var typeDefs = [''];

  var schemasText = schemaObj.schemas.join('\n');
  var queriesText = schemaObj.queryFields.join('\n');
  var mutationsText = schemaObj.mutationFields.join('\n');

  var queryTypeText = '';
  if (schemaObj.queryFields.length) {
    queryTypeText = '\n      type Query {\n        ' + queriesText + '\n      }\n    ';
  }

  var mutationTypeText = '';
  if (schemaObj.mutationFields.length) {
    mutationTypeText = '\n      type Mutation {\n        ' + mutationsText + '\n      }\n    ';
  }

  typeDefs[0] = '\n    scalar JSON\n    scalar Date\n    scalar DateTime\n\n    ' + schemasText + '\n\n    ' + queryTypeText + '\n\n    ' + mutationTypeText + '\n  ';

  var schema = makeExecutableSchema({
    typeDefs: typeDefs,
    resolvers: resolvers
  });

  schema._typeDefs = [print(parse(typeDefs[0]))];

  return schema;
}